import os
from functools import lru_cache
from pathlib import Path
from typing import Any

from dotenv import load_dotenv
from numpy.random import choice
from tqdm import trange

from cli import argument_parser
from models import Llama2, OpenAI
from task import QuestionAnswering, SentimentAnalysis
from utils import Prompt, log_calls, logger, save_snapshot


def conv2bool(_str: Any):
    if isinstance(_str, bool):
        return _str
    if str(_str).lower() in ["1", "true"]:
        return True
    if str(_str).lower() in ["0", "false"]:
        return False
    return None


load_dotenv()

current_directory = Path(__file__).resolve().parent

PARAPHRASE_PROMPT = """
Below is an instruction that describes a task. Write a response that paraphrases the instruction. Only output the paraphrased instruction bracketed in <prompt> and </prompt>.

### Instruction:
{instruction}

### Response:
<prompt>
"""


@log_calls("Paraphrasing prompts")
def paraphrase_prompts(prompt: str, n: int):
    paraphrases = []
    for _ in range(n):
        paraphrase = evolution_model(
            prompt=PARAPHRASE_PROMPT.format(instruction=prompt)
        )
        paraphrases.append(paraphrase)
    return paraphrases


@log_calls("Performing selection")
def selection(prompts):
    # In GA, two parent solutions are normally selected based on the roulette wheel
    # selection method according to the fitness value (Lipowski & Lipowska, 2012).
    # Similar to this, we utilize the roulette wheel selection method to select
    # two parent prompts in the current population according to the scores evaluated
    # on development sets. Specifically, let si denote the performance score on the
    # development set of the i-th prompt in the population, which contains a total
    # of N prompts. The probability of selecting the i-th prompt as a parent can be expressed as
    # pi = si / Σj=1->N sj.
    scores = [prompt.score for prompt in prompts]
    if sum(scores) == 0:
        # sum of scores is 0 ==> each score is 0, draw with equal probability
        selection_probabilities = len(scores) * [1 / len(scores)]
    else:
        selection_probabilities = [score / sum(scores) for score in scores]
    return choice(prompts, size=2, replace=False, p=selection_probabilities)


GA_PROMPT = """
Please follow the instruction step-by-step to generate a better prompt.
1. Cross over the following prompts and generate a new prompt:
Prompt 1: {prompt1}
Prompt 2: {prompt2}
2. Mutate the prompt generated in Step 1 and generate a final prompt bracketed with <prompt> and </prompt>.
"""


DE_PROMPT = """
Please follow the instruction step-by-step to generate a better prompt.
1. Identify the different parts between the Prompt 1 and Prompt 2:
Prompt 1: {prompt1}
Prompt 2: {prompt2}
2. Randomly mutate the different parts
3. Combine the different parts with Prompt 3, selectively replace it with the different parts in Step 2 and generate a new prompt.
Prompt 3: {prompt3}
4. Cross over the prompt in the Step 3 with the following basic prompt and generate a final prompt bracketed with <prompt> and </prompt>:
Basic Prompt: {basic_prompt}
"""


@log_calls("Performing prompt evolution using GA")
def evolution_ga(prompt1: str, prompt2: str):
    # Following the evolutionary operators in GA, a new candidate prompt is generated through
    # a two-step process based on the selected two parents:
    # 1) The parent prompts undergo crossover, resulting in a new prompt that
    #   selectively combines components from both parents;
    # 2) The newly generated prompt from the first step undergoes mutation,
    #   in which random alterations are made to some of its content.
    # Based on this two-step process, we design instructions, guiding LLMs to
    # generate a new prompt based on these steps to perform Evo(·) in Algorithm 1.
    evolved_prompt = evolution_model(
        prompt=GA_PROMPT.format(prompt1=prompt1, prompt2=prompt2)
    )
    if "<prompt>" in evolved_prompt:
        evolved_prompt = evolved_prompt.split("<prompt>")[1].split("</prompt>")[0]
    return evolved_prompt


@log_calls("Performing prompt evolution using DE")
def evolution_de(prompt1: str, prompt2: str, basic_prompt: str, best_prompt: str):
    # TODO add comment from paper
    evolved_prompt = evolution_model(
        prompt=DE_PROMPT.format(
            prompt1=prompt1,
            prompt2=prompt2,
            prompt3=best_prompt,
            basic_prompt=basic_prompt,
        )
    )
    if "<prompt>" in evolved_prompt:
        evolved_prompt = evolved_prompt.split("<prompt>")[1].split("</prompt>")[0]
    return evolved_prompt


@log_calls("Updating prompts")
def update(prompts: list[str], N: int):
    # EVOPROMPT iteratively generates new candidate prompts and assesses each prompt
    # using a development set, denoted as D, to obtain a score that quantifies the
    # quality of the prompt. We consider a straightforward selection strategy.
    # Specifically, at each iteration, EVOPROMPT based on GA produces N new prompts,
    # which are combined with the current population of N prompts.
    # The updated population is then selected by retaining the N prompts with the highest scores.
    retained_prompts: list[Prompt] = []
    min_retained_score = 0
    for prompt in prompts:
        if len(retained_prompts) < N:
            retained_prompts.append(prompt)
            min_retained_score = min(min_retained_score, prompt.score)
        elif prompt.score > min_retained_score:
            retained_prompts.sort(key=lambda p: p.score)
            retained_prompts[0] = prompt

    return retained_prompts


def run_episode(evo_alg_str: str, debug: bool = False):
    # Algorithm 1 Discrete prompt optimization: EVOPROMPT

    # Require:
    # - Size of population
    N = 3 if debug else 10
    # - Initial prompts P0 = {p1, p2, . . . , pN }
    paraphrases = paraphrase_prompts(task.base_prompt, n=N - 1)
    # the initial population
    initial_population = [task.base_prompt] + paraphrases

    # - fD(·) denotes the score of a prompt on the desired LLM evaluated on D
    f_D = lru_cache(maxsize=None)(task.evaluate_validation)

    # - a pre-defined number of iterations T
    T = 2 if debug else 10

    # - carefully designed evolutionary operators to generate a new prompt Evo(·)

    # Line 1: Initial evaluation scores: S0 ← {si = fD (pi )|i ∈ [1, N ]}
    # the current population's scores
    population_scores = [f_D(p) for p in initial_population]

    # all_prompts contains a list of Prompt objects that took part in this run at some time
    # converting prompts to Prompt object
    all_prompts: dict[str, Prompt] = {
        prompt.id: prompt
        for prompt in [
            Prompt(p, score=score, gen=0)
            for (p, score) in zip(initial_population, population_scores)
        ]
    }

    # P keeps track of prompts in each generation
    P = [[prompt_id for prompt_id in all_prompts.keys()]]

    # add initial prompts to family tree
    # None marks that there is no parent
    family_tree: dict[str, tuple[str, str] | None] = {
        prompt_id: None for prompt_id in P[0]
    }

    # Line 2:
    for t in trange(1, T + 1, desc="T", leave=True):
        # Line 3: Selection: select a certain number of prompts from current population as parent prompts
        # pr1,...,prk ∼ Pt−1
        prompts_current_evolution = [all_prompts[prompt_id] for prompt_id in P[t - 1]]
        if evo_alg_str == "de":
            # DE needs best prompt for evolution
            best_prompt_current_evolution = max(
                prompts_current_evolution, key=lambda prompt: prompt.score
            )

        new_evolutions = []

        for i in trange(N, desc="N", leave=False):
            # for both GA and DE we start with two parent prompts
            pr1, pr2 = selection([all_prompts[prompt_id] for prompt_id in P[t - 1]])

            # Line 4: Evolution: generate a new prompt based on the selected parent prompts by leveraging LLM to perform evolutionary operators
            # p′i ←Evo(pr1,...,prk)
            if evo_alg_str == "ga":
                p_i = evolution_ga(pr1, pr2)
            elif evo_alg_str == "de":
                p_i = evolution_de(
                    pr1,
                    pr2,
                    prompts_current_evolution[i],
                    best_prompt_current_evolution,
                )

            # Line 5: Evaluation
            # s′_i ← f(p′i,D)
            s_i = f_D(p_i)

            evolved_prompt = Prompt(content=p_i, score=s_i, gen=t)

            # keep track of genealogy
            family_tree[evolved_prompt.id] = (pr1.id, pr2.id)

            new_evolutions.append(evolved_prompt)
        all_prompts |= {prompt.id: prompt for prompt in new_evolutions}
        # Line 6: Update based on the evaluation scores
        # Pt ← {Pt−1, p′i} and St ← {St−1, s′i}
        if evo_alg_str == "ga":
            # GA keeps N best prompts from current population and evolutions
            population = update(new_evolutions + prompts_current_evolution, N)
        elif evo_alg_str == "de":
            # for DE we keep the evolved prompt if it is better than the basic prompt, and use the basic prompt otherwise
            assert len(prompts_current_evolution) == len(new_evolutions)
            population = [
                (
                    new_prompt
                    if new_prompt.score > current_prompt.score
                    else current_prompt
                )
                for current_prompt, new_prompt in zip(
                    prompts_current_evolution, new_evolutions
                )
            ]

        # store new generation
        P.append([prompt.id for prompt in population])

    save_snapshot(
        all_prompts,
        family_tree,
        P,
        T,
        N,
        task,
        evolution_model,
        options.__dict__,
    )
    # Line 8: Return the best prompt, p∗, among the final population PT :
    # p∗ ← argmaxp∈PT f(p, D)
    p = all_prompts[max(P[-1], key=lambda prompt_id: all_prompts[prompt_id].score)]
    logger.info(f"Best prompt: {p}")

    # We pick the prompt with the highest score on the development set and report its score on the testset.
    test_performance = task.evaluate_test(p.content)
    logger.info(f"Best prompt on test set: {test_performance}")


if __name__ == "__main__":
    options = argument_parser.parse_args()
    # log cli arguments
    logger.info(
        "CLI arguments:\n\tPositional:%s\n\tKeyword:\n\t\t%s",
        ", ".join(options._get_args()),
        "\n\t\t".join((f"{param}={value}" for param, value in options._get_kwargs())),
    )

    # debug mode will allow for a quick run
    debug = options.debug
    if debug is None:
        debug = conv2bool(os.getenv("EP_DEBUG", False))
        if debug is None:
            raise ValueError(
                f"{os.getenv('EP_DEBUG')} is not allowed for env variable EP_DEBUG."
            )

    # set up evolution model
    match options.evolution_engine:
        case "llama2":
            logger.info("Using Llama2 client as the evolution engine")
            evolution_model = Llama2(
                model_path=options.model_path,
                chat=options.chat,
            )

        case "openai":
            logger.info("Using OpenAI client as the evolution engine")
            evolution_model = OpenAI("gpt-3.5-turbo", chat=options.chat)

    # set up evaluation model
    # NOTE currenty we always stick to Llama2 as evaluation model
    match options.evolution_engine:
        case "llama2":
            evaluation_model = evolution_model
        case "openai":
            evaluation_model = Llama2(
                model_path=options.model_path,
                chat=options.chat,
            )

    match options.task:
        case "sa":
            logger.info("Running with task sentiment analysis on dataset SetFit/sst2")
            task = SentimentAnalysis(
                evaluation_model,
                "SetFit/sst2",
                "SetFit/sst2",
                validation_split=f"validation[:{5 if debug else 200}]",
                test_split="test[:20]" if debug else "test",
            )
        case "qa":
            logger.info("Running with task question answering on dataset squad")
            task = QuestionAnswering(
                evaluation_model,
                "squad",
                "squad",
                validation_split=f"train[:{5 if debug else 200}]",
                test_split="validation[:20]" if debug else "validation",
            )
        case _:
            raise ValueError(
                f"Task {options.task} does not exist. Choose from 'sa', 'qa'."
            )

    logger.info("Using evolutionary algorithm '%s'", options.evolution_algorithm)

    run_episode(evo_alg_str=options.evolution_algorithm, debug=debug)
